@@ -25,6 +25,9 @@
  *           The University of New South Wales
  *           Sydney, 2052
  *           Australia
+ *
+ *    Additions for bitmap and write-behind RAID options, Copyright (C) 2003-2004, 
+ *    Paul Clements, SteelEye Technology, Inc.
  */
 
 #include "mdadm.h"
@@ -47,6 +50,7 @@ int main(int argc, char *argv[])
 	int level = UnSet;
 	int layout = UnSet;
 	int raiddisks = 0;
+	int max_disks = MD_SB_DISKS;
 	int sparedisks = 0;
 	struct mddev_ident_s ident;
 	char *configfile = NULL;
@@ -56,17 +60,29 @@ int main(int argc, char *argv[])
 	char devmode = 0;
 	int runstop = 0;
 	int readonly = 0;
+	int write_behind = 0;
+	int bitmap_fd = -1;
+	char *bitmap_file = NULL;
+	int bitmap_chunk = UnSet;
 	int SparcAdjust = 0;
 	mddev_dev_t devlist = NULL;
 	mddev_dev_t *devlistend = & devlist;
 	mddev_dev_t dv;
 	int devs_found = 0;
 	int verbose = 0;
+	int quiet = 0;
 	int brief = 0;
 	int force = 0;
 	int test = 0;
 	int assume_clean = 0;
-	int autof = 0; /* -1 for non-partitions, 1 or more to create partitions */
+	int autof = 0; /* -2 means create device based on name:
+			*    if it ends mdN, then non-partitioned array N
+			*    if it ends dN, then partitions array N
+			* -1 means create non-partitioned, choose N
+			*  1 or more to create partitioned
+			* If -1 or 1 and name is a 'standard' name, then
+			* insist on a match of type and number.
+			*/
 
 	char *mailaddr = NULL;
 	char *program = NULL;
@@ -75,23 +91,34 @@ int main(int argc, char *argv[])
 	char *pidfile = NULL;
 	int oneshot = 0;
 	struct supertype *ss = NULL;
+	int writemostly = 0;
+	int re_add = 0;
+	char *shortopt = short_options;
+	int dosyslog = 0;
 
 	int copies;
 
 	int mdfd = -1;
 
+	srandom(time(0) ^ getpid());
+
 	ident.uuid_set=0;
 	ident.level = UnSet;
 	ident.raid_disks = UnSet;
 	ident.super_minor= UnSet;
 	ident.devices=0;
+	ident.spare_group = NULL;
+	ident.autof = 0;
+	ident.st = NULL;
+	ident.bitmap_fd = -1;
+	ident.name[0] = 0;
 
 	while ((option_index = -1) ,
 	       (opt=getopt_long(argc, argv,
-				short_options, long_options,
+				shortopt, long_options,
 				&option_index)) != -1) {
 		int newmode = mode;
-		/* firstly, so mode-independant options */
+		/* firstly, some mode-independant options */
 		switch(opt) {
 		case 'h':
 			help_text = Help;
@@ -115,10 +142,20 @@ int main(int argc, char *argv[])
 			fputs(Version, stderr);
 			exit(0);
 
-		case 'v': verbose = 1;
+		case 'v': verbose++;
 			continue;
 
-		case 'b': brief = 1;
+		case 'q': quiet++;
+			continue;
+
+		case 'b':
+			if (mode == ASSEMBLE || mode == BUILD || mode == CREATE || mode == GROW)
+				break; /* b means bitmap */
+			brief = 1;
+			if (optarg) {
+				fprintf(stderr, Name ": -b cannot have any extra immediately after it, sorry.\n");
+				exit(2);
+			}
 			continue;
 
 		case ':':
@@ -133,21 +170,29 @@ int main(int argc, char *argv[])
 
 		switch(opt) {
 		case '@': /* just incase they say --manage */
-			newmode = MANAGE; break;
+			newmode = MANAGE;
+			shortopt = short_bitmap_auto_options;
+			break;
 		case 'a':
 		case 'r':
 		case 'f':
-		case 1 : if (!mode) newmode = MANAGE; break;
+		case 6: /* re-add */
+			if (!mode) {
+				newmode = MANAGE;
+				shortopt = short_bitmap_auto_options;
+			}
+			break;
 
-		case 'A': newmode = ASSEMBLE; break;
-		case 'B': newmode = BUILD; break;
-		case 'C': newmode = CREATE; break;
+		case 'A': newmode = ASSEMBLE; shortopt = short_bitmap_auto_options; break;
+		case 'B': newmode = BUILD; shortopt = short_bitmap_auto_options; break;
+		case 'C': newmode = CREATE; shortopt = short_bitmap_auto_options; break;
 		case 'F': newmode = MONITOR;break;
-		case 'G': newmode = GROW; break;
+		case 'G': newmode = GROW; shortopt = short_bitmap_auto_options; break;
 
 		case '#':
 		case 'D':
 		case 'E':
+		case 'X':
 		case 'Q': newmode = MISC; break;
 		case 'R':
 		case 'S':
@@ -164,7 +209,7 @@ int main(int argc, char *argv[])
 				fprintf(stderr, "--%s", long_options[option_index].name);
 			else
 				fprintf(stderr, "-%c", opt);
-			fprintf(stderr, " would set mode to %s, but it is already %s.\n",
+			fprintf(stderr, " would set mdadm mode to \"%s\", but it is already set to \"%s\".\n",
 				map_num(modes, newmode),
 				map_num(modes, mode));
 			exit(2);
@@ -178,11 +223,35 @@ int main(int argc, char *argv[])
 				fputs(Help_config, stderr);
 				exit(0);
 			}
+
+			/* If first option is a device, don't force the mode yet */
+			if (opt == 1) {
+				if (devs_found == 0) {
+					dv = malloc(sizeof(*dv));
+					if (dv == NULL) {
+						fprintf(stderr, Name ": malloc failed\n");
+						exit(3);
+					}
+					dv->devname = optarg;
+					dv->disposition = devmode;
+					dv->writemostly = writemostly;
+					dv->re_add = re_add;
+					dv->next = NULL;
+					*devlistend = dv;
+					devlistend = &dv->next;
+			
+					devs_found++;
+					continue;
+				}
+				/* No mode yet, and this is the second device ... */
+				fprintf(stderr, Name ": An option must be given to set the mode before a second device is listed\n");
+				exit(2);
+			}
 			if (option_index >= 0)
-				fprintf(stderr, "--%s", long_options[option_index].name);
+				fprintf(stderr, Name ": --%s", long_options[option_index].name);
 			else
-				fprintf(stderr, "-%c", opt);
-			fprintf(stderr, " does not set the mode, and so cannot be first.\n");
+				fprintf(stderr, Name ": -%c", opt);
+			fprintf(stderr, " does not set the mode, and so cannot be the first option.\n");
 			exit(2);
 		}
 
@@ -215,6 +284,8 @@ int main(int argc, char *argv[])
 			}
 			dv->devname = optarg;
 			dv->disposition = devmode;
+			dv->writemostly = writemostly;
+			dv->re_add = re_add;
 			dv->next = NULL;
 			*devlistend = dv;
 			devlistend = &dv->next;
@@ -256,6 +327,14 @@ int main(int argc, char *argv[])
 				fprintf(stderr, Name ": unrecognised metadata identifier: %s\n", optarg);
 				exit(2);
 			}
+			max_disks = ss->max_devs;
+			continue;
+
+		case O(MANAGE,'W'):
+		case O(BUILD,'W'):
+		case O(CREATE,'W'):
+			/* set write-mostly for following devices */
+			writemostly = 1;
 			continue;
 
 		case O(GROW,'z'):
@@ -379,17 +458,11 @@ int main(int argc, char *argv[])
 				exit(2);
 			}
 			raiddisks = strtol(optarg, &c, 10);
-			if (!optarg[0] || *c || raiddisks<=0 || raiddisks > MD_SB_DISKS) {
+			if (!optarg[0] || *c || raiddisks<=0) {
 				fprintf(stderr, Name ": invalid number of raid devices: %s\n",
 					optarg);
 				exit(2);
 			}
-			if (raiddisks == 1 &&  !force && level != -5) {
-				fprintf(stderr, Name ": '1' is an unusual number of drives for an array, so it is probably\n"
-					"     a mistake.  If you really mean it you will need to specify --force before\n"
-					"     setting the number of drives.\n");
-				exit(2);
-			}
 			ident.raid_disks = raiddisks;
 			continue;
 
@@ -405,7 +478,7 @@ int main(int argc, char *argv[])
 				exit(2);
 			}
 			sparedisks = strtol(optarg, &c, 10);
-			if (!optarg[0] || *c || sparedisks < 0 || sparedisks > MD_SB_DISKS - raiddisks) {
+			if (!optarg[0] || *c || sparedisks < 0) {
 				fprintf(stderr, Name ": invalid number of spare-devices: %s\n",
 					optarg);
 				exit(2);
@@ -416,10 +489,12 @@ int main(int argc, char *argv[])
 		case O(BUILD,'a'):
 		case O(ASSEMBLE,'a'): /* auto-creation of device node */
 			if (optarg == NULL)
-				autof = -1;
+				autof = -2;
 			else if (strcasecmp(optarg,"no")==0)
 				autof = 0;
-			else if (strcasecmp(optarg,"yes")==0 || strcasecmp(optarg,"md")==0)
+			else if (strcasecmp(optarg,"yes")==0)
+				autof = -2;
+			else if (strcasecmp(optarg,"md")==0)
 				autof = -1;
 			else {
 				/* There might be digits, and maybe a hypen, at the end */
@@ -449,6 +524,7 @@ int main(int argc, char *argv[])
 			continue;
 
 		case O(BUILD,'f'): /* force honouring '-n 1' */
+		case O(GROW,'f'): /* ditto */
 		case O(CREATE,'f'): /* force honouring of device list */
 		case O(ASSEMBLE,'f'): /* force assembly */
 		case O(MISC,'f'): /* force zero */
@@ -470,6 +546,21 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
+		case O(CREATE,'N'):
+		case O(ASSEMBLE,'N'):
+			if (ident.name[0]) {
+				fprintf(stderr, Name ": name cannot be set twice.   "
+					"Second value %s.\n", optarg);
+				exit(2);
+			}
+			if (strlen(optarg) > 32) {
+				fprintf(stderr, Name ": name '%s' is too long, 32 chars max.\n",
+					optarg);
+				exit(2);
+			}
+			strcpy(ident.name, optarg);
+			continue;
+
 		case O(ASSEMBLE,'m'): /* super-minor for array */
 			if (ident.super_minor != UnSet) {
 				fprintf(stderr, Name ": super-minor cannot be set twice.  "
@@ -502,7 +593,23 @@ int main(int argc, char *argv[])
 				continue;
 			if (strcmp(update, "resync")==0)
 				continue;
-			fprintf(stderr, Name ": '--update %s' invalid.  Only 'sparc2.2', 'super-minor', 'resync' or 'summaries' supported\n",update);
+			if (strcmp(update, "uuid")==0)
+				continue;
+			if (strcmp(update, "byteorder")==0) {
+				if (ss) {
+					fprintf(stderr, Name ": must not set metadata type with --update=byteorder.\n");
+					exit(2);
+				}
+				for(i=0; !ss && superlist[i]; i++) 
+					ss = superlist[i]->match_metadata_desc("0.swap");
+				if (!ss) {
+					fprintf(stderr, Name ": INTERNAL ERROR cannot find 0.swap\n");
+					exit(2);
+				}
+
+				continue;
+			}
+			fprintf(stderr, Name ": '--update %s' invalid.  Only 'sparc2.2', 'super-minor', 'uuid', 'resync' or 'summaries' supported\n",update);
 			exit(2);
 
 		case O(ASSEMBLE,'c'): /* config file */
@@ -539,6 +646,9 @@ int main(int argc, char *argv[])
 			continue;
 
 		case O(MONITOR,'d'): /* delay in seconds */
+		case O(GROW, 'd'):
+		case O(BUILD,'d'): /* delay for bitmap updates */
+		case O(CREATE,'d'):
 			if (delay)
 				fprintf(stderr, Name ": only specify delay once. %s ignored.\n",
 					optarg);
@@ -567,6 +677,10 @@ int main(int argc, char *argv[])
 		case O(MONITOR,'t'): /* test */
 			test = 1;
 			continue;
+		case O(MONITOR,'y'): /* log messages to syslog */
+			openlog("mdadm", 0, SYSLOG_FACILITY);
+			dosyslog = 1;
+			continue;
 
 			/* now the general management options.  Some are applicable
 			 * to other modes. None have arguments.
@@ -574,6 +688,11 @@ int main(int argc, char *argv[])
 		case O(GROW,'a'):
 		case O(MANAGE,'a'): /* add a drive */
 			devmode = 'a';
+			re_add = 0;
+			continue;
+		case O(MANAGE,6):
+			devmode = 'a';
+			re_add = 1;
 			continue;
 		case O(MANAGE,'r'): /* remove a drive */
 			devmode = 'r';
@@ -620,6 +739,7 @@ int main(int argc, char *argv[])
 		case O(MISC,'K'):
 		case O(MISC,'R'):
 		case O(MISC,'S'):
+		case O(MISC,'X'):
 		case O(MISC,'o'):
 		case O(MISC,'w'):
 			if (devmode && devmode != opt &&
@@ -641,6 +761,56 @@ int main(int argc, char *argv[])
 			}
 			SparcAdjust = 1;
 			continue;
+
+		case O(ASSEMBLE,'b'): /* here we simply set the bitmap file */
+			if (!optarg) {
+				fprintf(stderr, Name ": bitmap file needed with -b in --assemble mode\n");
+				exit(2);
+			}
+			if (strcmp(optarg, "internal")==0) {
+				fprintf(stderr, Name ": there is no need to specify --bitmap when assembling arrays with internal bitmaps\n");
+				continue;
+			}
+			bitmap_fd = open(optarg, O_RDWR);
+			if (!*optarg || bitmap_fd < 0) {
+				fprintf(stderr, Name ": cannot open bitmap file %s: %s\n", optarg, strerror(errno));
+				exit(2);
+			}
+			ident.bitmap_fd = bitmap_fd; /* for Assemble */
+			continue;
+
+		case O(GROW,'b'):
+		case O(BUILD,'b'):
+		case O(CREATE,'b'): /* here we create the bitmap */
+			bitmap_file = optarg;
+			continue;
+
+		case O(GROW,4):
+		case O(BUILD,4):
+		case O(CREATE,4): /* bitmap chunksize */
+			bitmap_chunk = strtol(optarg, &c, 10);
+			if (!optarg[0] || *c || bitmap_chunk < 0 ||
+					bitmap_chunk & (bitmap_chunk - 1)) {
+				fprintf(stderr, Name ": invalid bitmap chunksize: %s\n",
+						optarg);
+				exit(2);
+			}
+			/* convert K to B, chunk of 0K means 512B */
+			bitmap_chunk = bitmap_chunk ? bitmap_chunk * 1024 : 512;
+			continue;
+
+		case O(BUILD, 5):
+		case O(CREATE, 5): /* write-behind mode */
+			write_behind = DEFAULT_MAX_WRITE_BEHIND;
+			if (optarg) {
+				write_behind = strtol(optarg, &c, 10);
+				if (write_behind < 0 || *c ||
+				    write_behind > 16383) {
+					fprintf(stderr, Name ": Invalid value for maximum outstanding write-behind writes: %s.\n\tMust be between 0 and 16383.\n", optarg);
+					exit(2);
+				}
+			}
+			continue;
 		}
 		/* We have now processed all the valid options. Anything else is
 		 * an error
@@ -651,6 +821,12 @@ int main(int argc, char *argv[])
 
 	}
 
+	if (!mode && devs_found) {
+		mode = MISC;
+		devmode = 'Q';
+		if (devlist->disposition == 0)
+			devlist->disposition = devmode;
+	}
 	if (!mode) {
 		fputs(Usage, stderr);
 		exit(2);
@@ -681,7 +857,28 @@ int main(int argc, char *argv[])
 		if ((int)ident.super_minor == -2) {
 			struct stat stb;
 			fstat(mdfd, &stb);
-			ident.super_minor = MINOR(stb.st_rdev);
+			ident.super_minor = minor(stb.st_rdev);
+		}
+	}
+
+	if (raiddisks) {
+		if (raiddisks > max_disks) {
+			fprintf(stderr, Name ": invalid number of raid devices: %d\n",
+				raiddisks);
+			exit(2);
+		}
+		if (raiddisks == 1 &&  !force && level != -5) {
+			fprintf(stderr, Name ": '1' is an unusual number of drives for an array, so it is probably\n"
+				"     a mistake.  If you really mean it you will need to specify --force before\n"
+				"     setting the number of drives.\n");
+			exit(2);
+		}
+	}
+	if (sparedisks) {
+		if ( sparedisks > max_disks - raiddisks) {
+			fprintf(stderr, Name ": invalid number of spare-devices: %d\n",
+				sparedisks);
+			exit(2);
 		}
 	}
 
@@ -693,15 +890,15 @@ int main(int argc, char *argv[])
 			rv = Manage_ro(devlist->devname, mdfd, readonly);
 		if (!rv && devs_found>1)
 			rv = Manage_subdevs(devlist->devname, mdfd,
-					    devlist->next);
+					    devlist->next, verbose-quiet);
 		if (!rv && readonly < 0)
 			rv = Manage_ro(devlist->devname, mdfd, readonly);
 		if (!rv && runstop)
-			rv = Manage_runstop(devlist->devname, mdfd, runstop);
+			rv = Manage_runstop(devlist->devname, mdfd, runstop, 0);
 		break;
 	case ASSEMBLE:
 		if (devs_found == 1 && ident.uuid_set == 0 &&
-		    ident.super_minor == UnSet && !scan ) {
+		    ident.super_minor == UnSet && ident.name[0] == 0 && !scan ) {
 			/* Only a device has been given, so get details from config file */
 			mddev_ident_t array_ident = conf_get_ident(configfile, devlist->devname);
 			if (array_ident == NULL) {
@@ -709,20 +906,21 @@ int main(int argc, char *argv[])
 					devlist->devname);
 				rv |= 1;
 			} else {
-				mdfd = open_mddev(devlist->devname, array_ident->autof);
+				mdfd = open_mddev(devlist->devname, 
+						  array_ident->autof ? array_ident->autof : autof);
 				if (mdfd < 0)
 					rv |= 1;
 				else {
 					rv |= Assemble(ss, devlist->devname, mdfd, array_ident, configfile,
 						       NULL,
-						       readonly, runstop, update, verbose, force);
+						       readonly, runstop, update, verbose-quiet, force);
 					close(mdfd);
 				}
 			}
 		} else if (!scan)
 			rv = Assemble(ss, devlist->devname, mdfd, &ident, configfile,
 				      devlist->next,
-				      readonly, runstop, update, verbose, force);
+				      readonly, runstop, update, verbose-quiet, force);
 		else if (devs_found>0) {
 			if (update && devs_found > 1) {
 				fprintf(stderr, Name ": can only update a single array at a time\n");
@@ -736,14 +934,15 @@ int main(int argc, char *argv[])
 					rv |= 1;
 					continue;
 				}
-				mdfd = open_mddev(dv->devname, array_ident->autof);
+				mdfd = open_mddev(dv->devname, 
+						  array_ident->autof ?array_ident->autof : autof);
 				if (mdfd < 0) {
 					rv |= 1;
 					continue;
 				}
 				rv |= Assemble(ss, dv->devname, mdfd, array_ident, configfile,
 					       NULL,
-					       readonly, runstop, update, verbose, force);
+					       readonly, runstop, update, verbose-quiet, force);
 				close(mdfd);
 			}
 		} else {
@@ -754,25 +953,51 @@ int main(int argc, char *argv[])
 			} else
 				for (; array_list; array_list = array_list->next) {
 					mdu_array_info_t array;
-					mdfd = open_mddev(array_list->devname, array_list->autof);
+					mdfd = open_mddev(array_list->devname, 
+							  array_list->autof ? array_list->autof : autof);
 					if (mdfd < 0) {
 						rv |= 1;
 						continue;
 					}
 					if (ioctl(mdfd, GET_ARRAY_INFO, &array)>=0)
 						/* already assembled, skip */
-						continue;
-					rv |= Assemble(ss, array_list->devname, mdfd,
-						       array_list, configfile,
-						       NULL,
-						       readonly, runstop, NULL, verbose, force);
+						;
+					else
+						rv |= Assemble(ss, array_list->devname, mdfd,
+							       array_list, configfile,
+							       NULL,
+							       readonly, runstop, NULL, verbose-quiet, force);
+					close(mdfd);
 				}
 		}
 		break;
 	case BUILD:
-		rv = Build(devlist->devname, mdfd, chunk, level, layout, raiddisks, devlist->next, assume_clean);
+		if (bitmap_chunk == UnSet) bitmap_chunk = DEFAULT_BITMAP_CHUNK;
+		if (delay == 0) delay = DEFAULT_BITMAP_DELAY;
+		if (write_behind && !bitmap_file) {
+			fprintf(stderr, Name ": write-behind mode requires a bitmap.\n");
+			rv = 1;
+			break;
+		}
+
+		if (bitmap_file) {
+			if (strcmp(bitmap_file, "internal")==0) {
+				fprintf(stderr, Name ": 'internal' bitmaps not supported with --build\n");
+				rv |= 1;
+				break;
+			}
+		}
+		rv = Build(devlist->devname, mdfd, chunk, level, layout,
+			   raiddisks, devlist->next, assume_clean,
+			   bitmap_file, bitmap_chunk, write_behind, delay, verbose-quiet);
 		break;
 	case CREATE:
+		if (delay == 0) delay = DEFAULT_BITMAP_DELAY;
+		if (write_behind && !bitmap_file) {
+			fprintf(stderr, Name ": write-behind mode requires a bitmap.\n");
+			rv = 1;
+			break;
+		}
 		if (ss == NULL) {
 			for(i=0; !ss && superlist[i]; i++) 
 				ss = superlist[i]->match_metadata_desc("default");
@@ -783,8 +1008,9 @@ int main(int argc, char *argv[])
 		}
 
 		rv = Create(ss, devlist->devname, mdfd, chunk, level, layout, size<0 ? 0 : size,
-			    raiddisks, sparedisks,
-			    devs_found-1, devlist->next, runstop, verbose, force);
+			    raiddisks, sparedisks, ident.name,
+			    devs_found-1, devlist->next, runstop, verbose-quiet, force, assume_clean,
+			    bitmap_file, bitmap_chunk, write_behind, delay);
 		break;
 	case MISC:
 
@@ -799,11 +1025,11 @@ int main(int argc, char *argv[])
 				fprintf(stderr, Name ": No devices listed in %s\n", configfile?configfile:DefaultConfFile);
 				exit(1);
 			}
-			rv = Examine(devlist, scan?!verbose:brief, scan, SparcAdjust, ss);
+			rv = Examine(devlist, scan?(verbose>1?0:verbose+1):brief, scan, SparcAdjust, ss);
 		} else {
 			if (devlist == NULL) {
-				if ((devmode == 'S' ||devmode=='D') && scan) {
-					/* apply to all devices in /proc/mdstat */
+				if (devmode=='D' && scan) {
+					/* apply --detail to all devices in /proc/mdstat */
 					struct mdstat_ent *ms = mdstat_read(0);
 					struct mdstat_ent *e;
 					for (e=ms ; e ; e=e->next) {
@@ -814,16 +1040,44 @@ int main(int argc, char *argv[])
 								e->dev);
 							continue;
 						}
-						if (devmode == 'D')
-							rv |= Detail(name, !verbose, test);
-						else if (devmode=='S') {
-							mdfd = open_mddev(name, 0);
-							if (mdfd >= 0)
-								rv |= Manage_runstop(name, mdfd, -1);
-						}
+						rv |= Detail(name, verbose>1?0:verbose+1, test);
 						put_md_name(name);
 					}
-				} else {						
+				} else	if (devmode == 'S' && scan) {
+					/* apply --stop to all devices in /proc/mdstat */
+					/* Due to possible stacking of devices, repeat until
+					 * nothing more can be stopped
+					 */
+					int progress=1, err;
+					int last = 0;
+					do {
+						struct mdstat_ent *ms = mdstat_read(0);
+						struct mdstat_ent *e;
+
+						if (!progress) last = 1;
+						progress = 0; err = 0;
+						for (e=ms ; e ; e=e->next) {
+							char *name = get_md_name(e->devnum);
+
+							if (!name) {
+								fprintf(stderr, Name ": cannot find device file for %s\n",
+									e->dev);
+								continue;
+							}
+							mdfd = open_mddev(name, 0);
+							if (mdfd >= 0) {
+								if (Manage_runstop(name, mdfd, -1, !last))
+									err = 1;
+								else
+									progress = 1;
+								close(mdfd);
+							}
+
+							put_md_name(name);
+						}
+					} while (!last && err);
+					if (err) rv |= 1;
+				} else {
 					fprintf(stderr, Name ": No devices given.\n");
 					exit(2);
 				}
@@ -831,24 +1085,29 @@ int main(int argc, char *argv[])
 			for (dv=devlist ; dv; dv=dv->next) {
 				switch(dv->disposition) {
 				case 'D':
-					rv |= Detail(dv->devname, brief, test); continue;
+					rv |= Detail(dv->devname, brief?1+verbose:0, test); continue;
 				case 'K': /* Zero superblock */
 					rv |= Kill(dv->devname, force); continue;
 				case 'Q':
 					rv |= Query(dv->devname); continue;
+				case 'X':
+					rv |= ExamineBitmap(dv->devname, brief, ss); continue;
 				}
 				mdfd = open_mddev(dv->devname, 0);
-				if (mdfd>=0)
+				if (mdfd>=0) {
 					switch(dv->disposition) {
 					case 'R':
-						rv |= Manage_runstop(dv->devname, mdfd, 1); break;
+						rv |= Manage_runstop(dv->devname, mdfd, 1, 0); break;
 					case 'S':
-						rv |= Manage_runstop(dv->devname, mdfd, -1); break;
+						rv |= Manage_runstop(dv->devname, mdfd, -1, 0); break;
 					case 'o':
 						rv |= Manage_ro(dv->devname, mdfd, 1); break;
 					case 'w':
 						rv |= Manage_ro(dv->devname, mdfd, -1); break;
 					}
+					close(mdfd);
+				} else
+					rv |= 1;
 			}
 		}
 		break;
@@ -864,7 +1123,8 @@ int main(int argc, char *argv[])
 			break;
 		}
 		rv= Monitor(devlist, mailaddr, program,
-			    delay?delay:60, daemonise, scan, oneshot, configfile, test, pidfile);
+			    delay?delay:60, daemonise, scan, oneshot,
+			    dosyslog, configfile, test, pidfile);
 		break;
 
 	case GROW:
@@ -881,15 +1141,19 @@ int main(int argc, char *argv[])
 				if (rv)
 					break;
 			}
-		} else if ((size >= 0) + (raiddisks != 0) +  (layout != UnSet) > 1) {
-			fprintf(stderr, Name ": can change at most one of size, raiddisks, and layout\n");
+		} else if ((size >= 0) + (raiddisks != 0) +  (layout != UnSet) + (bitmap_file != NULL)> 1) {
+			fprintf(stderr, Name ": can change at most one of size, raiddisks, bitmap, and layout\n");
 			rv = 1;
 			break;
 		} else if (layout != UnSet)
 			rv = Manage_reconfig(devlist->devname, mdfd, layout);
 		else if (size >= 0 || raiddisks)
 			rv = Manage_resize(devlist->devname, mdfd, size, raiddisks);
-		else 
+		else if (bitmap_file) {
+			if (delay == 0) delay = DEFAULT_BITMAP_DELAY;
+			rv = Grow_addbitmap(devlist->devname, mdfd, bitmap_file,
+					    bitmap_chunk, delay, write_behind);
+		} else
 			fprintf(stderr, Name ": no changes to --grow\n");
 		break;
 	}
